package estimation;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;
import java.util.logging.Level;
import java.util.logging.Logger;

import cern.colt.matrix.DoubleMatrix2D;
import cern.colt.matrix.impl.DenseDoubleMatrix2D;
import cern.colt.matrix.linalg.Algebra;
import cern.jet.math.Functions;
import data.GrmReader;
import data.IGrmReader;
import data.IndividualVariables;
import data.SinglePcGrmReader;
import paramfile.ParameterSet;
import paramfile.parameters.IntegerParameter;
import paramfile.parameters.StringParameter;
import paramfile.parameters.ValidationFailedException;

public class ResidualizeGrm {
	private static final String GRM = "grm";
	private static final String MODCOUNT = "modcount";
	private static final String COVARS = "covars";
	
	
	private static Logger logger = Logger.getLogger(ResidualizeGrm.class.getName());
	static
	{
		logger.setLevel(Level.INFO);
	}
	private static ParameterSet getParameters(String filename) throws FileNotFoundException, IOException, ValidationFailedException {
		ParameterSet params = new ParameterSet();
		params.addRequiredParameter(new StringParameter(GRM));
		params.addRequiredParameter(new StringParameter(COVARS));
		params.addRequiredParameter(new IntegerParameter(MODCOUNT, Integer.MAX_VALUE));
		params.load(filename);
		return params;
	}
	
	
	//starting with a random assignment of individuals to cohorts
	public static void main(String[] args) throws IOException, ClassNotFoundException, ValidationFailedException, InterruptedException, URISyntaxException
	{
		ParameterSet p = getParameters(args[0]);
		String prefix = p.getStringValue(GRM);
		String idfile = prefix + ".grm.id";
		String binfile = prefix + ".grm.bin";
		GrmReader grm_reader = new GrmReader(binfile, idfile);
		
		IndividualVariables ids = grm_reader.getIds();
		
		String covars = p.getStringValue(COVARS);
		int modcount = p.getIntValue(MODCOUNT);
		logger.log(Level.INFO, "reading GRMS");
		List<IGrmReader> l = new ArrayList<IGrmReader>();
		getCovarGrms(covars, ids, l);
		l.add(grm_reader);

		DoubleMatrix2D den = new DenseDoubleMatrix2D(l.size(), l.size());
		DoubleMatrix2D num = new DenseDoubleMatrix2D(l.size(), 1);
		
		int count = 0;
		while(l.get(0).hasNext())
		{
			count += addNext(l, den, num);
			if((count % modcount == 0))
			{
				System.out.print(count + " pairs analyzed\r");
			}
		}
		print(den);
		print(num);
		double[] coefficients = estimate(num, den);
		for(int i = 0; i < coefficients.length; i++)
		{
			System.out.println(coefficients[i]);
		}

	}
	
	public static void print(DoubleMatrix2D mat)
	{
		System.out.println();
		for(int i = 0; i < mat.rows(); i++)
		{
			for(int j = 0; j < mat.columns(); j++)
			{
				System.out.print(mat.get(i, j) + "\t");
			}
			System.out.println();
		}
		System.out.println();
	}
	
	private static double[] estimate(DoubleMatrix2D num, DoubleMatrix2D den)
	{
		Algebra a = new Algebra();
		DoubleMatrix2D inv = a.inverse(den);
		DoubleMatrix2D coeffs = a.mult(inv, num);
		double[] results = new double[num.rows()];
		for(int i = 0; i < coeffs.rows(); i++)
		{
			double cg = coeffs.getQuick(i, 0);
			results[i] = cg;
		}
		return results;
	}
	private static int addNext(List<IGrmReader> l, DoubleMatrix2D den, DoubleMatrix2D num){
		double[][] vals = new double[l.size()][];
		double yval = -1;
		int index1 = 0, index2 = 0;
		boolean keep_looking = true;
		while(keep_looking)
		{
			if(!l.get(0).hasNext())
			{
				logger.log(Level.INFO, "no entries after " + l.get(0).getIndex1() + "," + l.get(0).getIndex2());
				return 0;
			}
			int i = 0;
			for(i = 0; i < l.size() - 1; i++)
			{
				IGrmReader s  = l.get(i);
				if(!s.hasNext())
				{
					throw new RuntimeException();
				}
				s.next();
				vals[i] = new double[1];
				vals[i][0] = s.getValue();
			}
			vals[i] = new double[1];
			vals[i][0] = 1;
			index1 = l.get(0).getIndex1();
			index2 = l.get(0).getIndex2();
			l.get(i).next();
			yval = l.get(i).getValue();
			keep_looking = false;
			if(index1 == index2)
			{
				keep_looking = true;
			}
		}
		DoubleMatrix2D d2d = new DenseDoubleMatrix2D(vals);
		d2d.zMult(d2d, den, 1, 1, false, true);
		num.assign(d2d, Functions.plusMult(yval));
		return 1;
	}


	public static int getCovarGrms(String covars_file, IndividualVariables ids, List<IGrmReader> l) throws IOException
	{
		if( covars_file != null )
		{
			IndividualVariables variables = IndividualVariables.read(covars_file);
//			l.add(new SinglePcGrmReader(ids, variables, 0));
			for(int vec = 0; vec < variables.getVariableCount(); vec++)
			{
				l.add(new SinglePcGrmReader(ids, variables, vec));
			}
		}
		return l.size();
	}
	
	public static List<String> getStrings(String listfile) throws IOException
	{
		Scanner scanner = null;
		List<String> l;
		try {
			scanner = new Scanner(new BufferedInputStream(new FileInputStream(listfile)));
			
			l = new ArrayList<String>();
			while(scanner.hasNext())
			{
				String s = scanner.next();
				l.add(s);
			}
		} finally
		{
			if( scanner != null )
			{
				scanner.close();
			}
		}
		return l;
	}
	
	
}
